# Test exporting fit parameters

from numpy.testing import *
import numpy as np
import pandas as pd
import bdata as bd
from bfit.gui.bfit import bfit
import os

# filter unneeded warnings
import pytest
pytestmark = pytest.mark.filterwarnings('ignore:2020')

# test file name
filename = 'TESTFILE.csv'

def with_bfit(function):
    
    def wrapper(*args, **kwargs):
        # make gui
        b = bfit(None, True)
        tab = b.fetch_files
        tab2 = b.fit_files
        b.notebook.select(2)
        b.draw_fit.set(False)
        
        # get data
        tab.year.set(2020)
        tab.run.set('40123 40127')
        tab.get_data()
        
        # fit
        tab2.populate()
        
        # test
        try:
            return function(*args, **kwargs, b=b)
        finally:
            b.on_closing()
            del b
            
    return wrapper

def check_columns(filename, columns):
    
    # load file
    df = pd.read_csv(filename, comment='#')
    
    # check columns
    for c in columns:
        assert c in df.columns, ('%s not included' % c)
    
    # check column length
    assert_equal(len(df.columns), len(columns))

@with_bfit
def test_columns_no_fit(b=None):
    
    b.fit_files.export(filename = filename)
    
    # get columns, no fit parameters (exp fit)
    columns = list(b.fit_files.xaxis_combobox['values'][3:])
    
    # add missing columns: errors
    new_columns = []
    for c in columns:
        
        # nmr values only
        if c in ['Cryo Lift Set (mm)',
                 'Cryo Lift Read (mm)']:
            continue
            
        # values not logged in this case
        if c in ['He Mass Flow',
                 'Needle Set (turns)',
                 'Needle Read (turns)',
                 ]:
            continue
            
        # no fit values
        if c in ['Chi-Squared',]:
            continue
        
        # copy values
        new_columns.append(c)
        
        # error-less quantities
        if c not in ['Unique Id',
                     'Run Number',
                     'Sample',
                     'Run Duration (s)',
                     'Start Time',
                     'Title',
                     'Year',
                     'NBM Rate (count/s)',
                     'Sample Rate (count/s)',
                     'Chi-Squared',
                     ]:
            new_columns.append('Error '+c)
        
    columns = new_columns
        
    # add missing columns: run number, shared, fixed
    for name in b.data['2020.40123'].parnames:
        columns.append('fixed '+name)
    
    check_columns(filename, columns)

    # clean
    os.remove(filename)
    
@with_bfit
def test_columns_with_fit(b=None):
    
    b.fit_files.do_fit()
    b.fit_files.export(filename = filename)
    
    # get columns, no fit parameters (exp fit)
    columns = list(b.fit_files.xaxis_combobox['values'][1:])
    
    # add missing columns: errors
    new_columns = []
    for c in columns:
        
        # nmr values only
        if c in ['Cryo Lift Set (mm)',
                 'Cryo Lift Read (mm)']:
            continue
            
        # values not logged in this case
        if c in ['He Mass Flow',
                 'Needle Set (turns)',
                 'Needle Read (turns)',
                 ]:
            continue
            
        # copy values
        new_columns.append(c)
        
        # error-less quantities
        if c not in ['Unique Id',
                     'Run Number',
                     'Sample',
                     'Run Duration (s)',
                     'Start Time',
                     'Title',
                     'Year',
                     'NBM Rate (count/s)',
                     'Sample Rate (count/s)',
                     'Chi-Squared',
                     ]:
                         
            if c in ['1_T1', 'amp']:
                new_columns.append('Error+ '+c)
                new_columns.append('Error- '+c)
            else:
                new_columns.append('Error '+c)
            
    columns = new_columns
        
    # add missing columns: run number, shared, fixed
    for name in b.data['2020.40123'].parnames:
        columns.append('fixed '+name)
        columns.append('shared '+name)
    
    check_columns(filename, columns)
    
    # clean
    os.remove(filename)

@with_bfit
def test_content(b=None):
    
    # set up and get data
    b.fit_files.do_fit()
    b.fit_files.export(filename = filename)
    df = pd.read_csv(filename, comment='#').loc[0]
    data = b.data['%d.%d' % (df['Year'], df['Run Number'])]
    
    # checker shortcut
    def chk(title, value):
        assert_almost_equal(df[title], value, err_msg='Compare %s'%title,
                            decimal=3)
    
    # check parameters
    chk('1_T1', data.fitpar.loc['1_T1', 'res'])
    chk('Error- 1_T1', data.fitpar.loc['1_T1', 'dres-'])
    chk('Error+ 1_T1', data.fitpar.loc['1_T1', 'dres+'])
    
    chk('amp', data.fitpar.loc['amp', 'res'])
    chk('Error- amp', data.fitpar.loc['amp', 'dres-'])
    chk('Error+ amp', data.fitpar.loc['amp', 'dres+'])
    
    # check parameters
    chk('Temperature (K)', data.temperature.mean)
    chk('Error Temperature (K)', data.temperature.std)
    
    chk('1000/T (1/K)', 1000/data.temperature.mean)
    chk('Error 1000/T (1/K)', 1000*data.temperature.std/data.temperature.mean**2)
    
    chk('Impl. Energy (keV)', data.beam_kev)
    chk('Error Impl. Energy (keV)', data.beam_kev_err)
    
    chk('Platform Bias (kV)', data.nmr_bias.mean)
    chk('Error Platform Bias (kV)', data.nmr_bias.std)
    
    chk('B0 Field (T)', data.field)
    chk('Error B0 Field (T)', data.field_std)
    
    assert df['Sample'] == data.sample, 'Sample'
    assert df['Title'] == data.title, 'Title'
    
    chk('RF Level DAC', data.rf_dac.mean)
    chk('Error RF Level DAC', data.rf_dac.std)
    
    chk('Chi-Squared', data.chi)
    
    chk('Run Duration (s)', data.duration)
    
    chk('Start Time', data.start_time)
    
    chk('Year', data.year)
    
    chk('CryoEx Mass Flow', data.cryo_read.mean)
    chk('Error CryoEx Mass Flow', data.cryo_read.std)
    
    chk('Laser Power', data.las_pwr.mean)
    chk('Error Laser Power', data.las_pwr.std)
    
    chk('Target Bias (kV)', data.target_bias.mean)
    chk('Error Target Bias (kV)', data.target_bias.std)
    
    chk('NBM Rate (count/s)', np.sum([data.hist['NBM'+c].data 
                                     for c in ['B+', 'B-', 'F+', 'F-']])/data.duration)
    chk('Sample Rate (count/s)', np.sum([data.hist[c].data 
                                     for c in ['B+', 'B-', 'F+', 'F-']])/data.duration)
    
    assert df['fixed 1_T1'] == data.fitpar.loc['1_T1', 'fixed'], 'fixed 1_T1'
    assert df['shared 1_T1'] == data.fitpar.loc['1_T1', 'shared'], 'shared 1_T1'
    assert df['fixed amp'] == data.fitpar.loc['amp', 'fixed'], 'fixed amp'
    assert df['shared amp'] ==data.fitpar.loc['amp', 'shared'], 'shared amp'
    
    # clean
    os.remove(filename)
    



# Test exporting asymmetries
from numpy.testing import *
import numpy as np
import pandas as pd
import bdata as bd
from bfit.gui.bfit import bfit
import os

# filter unneeded warnings
import pytest
pytestmark = pytest.mark.filterwarnings('ignore:2020')

# test file name
filename = 'TESTFILE.csv'

def with_bfit(function):
    
    def wrapper(*args, **kwargs):
        # make gui
        b = bfit(None, True)
        tab = b.fetch_files
        tab2 = b.fileviewer
        b.notebook.select(2)
        
        tab.year.set(2020)
        tab.run.set('40123 40127')
        tab.get_data()
        
        tab2.runn.set(40123)
        tab2.year.set(2020)
        tab2.get_data()
        
        try:
            return function(*args, **kwargs, b=b)
        finally:
            b.on_closing()
            del b
            
    return wrapper

def check_columns(filename):
    
    # load file
    df = pd.read_csv(filename, comment='#')
    
    # check columns
    columns = ['time_s', 
                'positive_helicity', 
                'positive_helicity_err',
                'negative_helicity', 
                'negative_helicity_err', 
                'forward_counter',
                'forward_counter_err', 
                'backward_counter', 
                'backward_counter_err',
                'combined', 
                'combined_err']
    for c in df.columns:
        assert c in columns
    
    # check column length
    assert len(df.columns) == len(columns)

def check_content(data, filename):
    
    # load file
    df = pd.read_csv(filename, comment='#')
    
    # get asymmetry
    asym = data.asym()
    
    # check time
    assert_array_almost_equal(df['time_s'],                asym['time_s'], err_msg='time compare')
    
    # check combined
    assert_array_almost_equal(df['combined'],              asym['c'][0], err_msg='combined compare')
    assert_array_almost_equal(df['combined_err'],          asym['c'][1], err_msg='combined error compare')
    
    # check split hel
    assert_array_almost_equal(df['positive_helicity'],     asym['p'][0], err_msg='positive helicity compare')
    assert_array_almost_equal(df['positive_helicity_err'], asym['p'][1], err_msg='positive helicity error compare')
    
    assert_array_almost_equal(df['negative_helicity'],     asym['n'][0], err_msg='negative helicity compare')
    assert_array_almost_equal(df['negative_helicity_err'], asym['n'][1], err_msg='negative helicity error compare')
    
    # check split detector
    assert_array_almost_equal(df['forward_counter'],       asym['f'][0], err_msg='forward counter compare')
    assert_array_almost_equal(df['forward_counter_err'],   asym['f'][1], err_msg='forward counter error compare')
    
    assert_array_almost_equal(df['backward_counter'],      asym['b'][0], err_msg='backward counter compare')
    assert_array_almost_equal(df['backward_counter_err'],  asym['b'][1], err_msg='backward counter error compare')

@with_bfit
def test_export_fileviewer_columns(b=None):
    
    # export
    b.fileviewer.export(filename)
    
    # test
    check_columns(filename)
    
    # clean
    os.remove(filename)

@with_bfit
def test_export_fileviewer_content(b=None):
    
    # export
    b.fileviewer.export(filename)
    
    # get bdata object
    data = b.fileviewer.data
    
    # test
    check_content(data, filename)
    
    # clean
    os.remove(filename)

@with_bfit
def test_export_fetch_files_columns(b=None):
    
    # export
    b.fetch_files.export(directory='.')
    
    # test
    check_columns('2020_40123.csv')
    check_columns('2020_40127.csv')
    
    # clean
    os.remove('2020_40123.csv')
    os.remove('2020_40127.csv')

@with_bfit
def test_export_fetch_files_content(b=None):
    
    # export
    b.fetch_files.export(directory='.')
    
    # get bdata object
    datalines = b.fetch_files.data_lines

    # test
    check_content(datalines['2020.40123'].bdfit, '2020_40123.csv')
    check_content(datalines['2020.40127'].bdfit, '2020_40127.csv')
    
    # clean
    os.remove('2020_40123.csv')
    os.remove('2020_40127.csv')
    



# Test fit functions
# Derek Fujimoto
# Feb 2021

from numpy.testing import *
from bfit.fitting.functions import *
import numpy as np

def test_lorentzian():
    
    # test values
    amp = 1
    fwhm = 0.25
    peak = 0.5
    
    gety = lambda x : -lorentzian(x, peak, fwhm, amp) # flip y for less confusion on my part
    dx = 1e-9   # window to check values in
        
    # check amplitude
    x = np.linspace(peak-dx/2, peak+dx/2, 10000)
    y = gety(x)
    assert_almost_equal(max(y), 1, err_msg = "lorentzian amplitude")
    
    # check peak location
    xmax = x[y == max(y)][0]
    assert_almost_equal(xmax, peak, err_msg = "lorentzian peak")
    
    # check FWHM
    xleft = np.linspace(peak-fwhm/2-dx, peak-fwhm/2+dx, 10000)
    xright = np.linspace(peak+fwhm/2-dx, peak+fwhm/2+dx, 10000)
    
    yleft = gety(xleft) - amp/2
    yright = gety(xright) - amp/2
    
    x1 = max(xleft[yleft < 0])
    x2 = min(xright[yright < 0])
    
    width = x2 - x1
    assert_almost_equal(width, fwhm, err_msg = "lorentzian fwhm")
    
def test_bilorentzian():
    
    x = np.linspace(0, 1, 10000)
    
    # test values for lor2 = 0
    amp = 1
    fwhm = 0.25
    peak = 0.5
    
    lor = lorentzian(x, peak, fwhm, amp)
    bi1 = bilorentzian(x, peak, fwhm, amp, 0, 0)
    bi2 = bilorentzian(x, peak, 0, 0, fwhm, amp)
    
    assert_array_almost_equal(lor, bi1, err_msg = "bilorenzian component 1")
    assert_array_almost_equal(lor, bi2, err_msg = "bilorenzian component 2")
    
def test_gaussian():
    
    # test values
    amp = 1
    sigma = 0.25
    mean = 0.5
    
    x = np.linspace(0, 1, 100000)
    y = -gaussian(x, mean, sigma, amp)
    
    # test amp
    assert_almost_equal(max(y), amp, err_msg = "gaussian amp")
    
    # test mean 
    avg = np.average(x, weights=y)
    assert_almost_equal(avg, mean, err_msg = "gaussian mean")
    
    # test stdev
    mean = 5
    x = np.linspace(0, 10, 100000)
    y = -gaussian(x, mean, sigma, amp)
    
    wsum = np.sum(np.square(mean - x) * y)
    sig = np.sqrt( wsum / np.sum(y) )
    assert_almost_equal(sig, sigma, err_msg = "gaussian sigma")
    
def test_quadlorentzian():
    
    # ~ nu_0 = 1e6
    nu_q = 1e3
    eta = 0
    theta = 0
    phi = 0
    
    # test first order peak locations for I = 2
    I = 2
    peaks = np.array([qp_1st_order(nu_q, eta, theta, phi, m) for m in np.arange(-(I-1), I+1, 1)])
    
    assert_almost_equal(len(peaks), 4, err_msg = "quadlorentzian number of first order peaks for I=2")
    assert_array_almost_equal(peaks, np.array([nu_q*3, nu_q, -nu_q, -3*nu_q]), 
                            err_msg = "quadlorentzian first order shifts for I=2")
    
    # test first order peak locations for I = 3/2
    I = 3/2
    peaks = np.array([qp_1st_order(nu_q, eta, theta, phi, m) for m in np.arange(-(I-1), I+1, 1)])
    assert_almost_equal(len(peaks), 3, err_msg = "quadlorentzian number of first order peaks for I=3/2")
    assert_array_almost_equal(peaks, np.array([nu_q*2, 0, -2*nu_q]), 
                                err_msg = "quadlorentzian first order shifts for I=3/2")
    
def test_pulsed_exp():
    
    # settings
    amp = 1
    tau = 1
    pulse_len = 4
    
    x = np.linspace(1e-9, 10, 10000)
    pexp = pulsed_exp(lifetime = tau, pulse_len = pulse_len)
    
    # test amp
    y = pexp(x, 1, amp)
    assert_almost_equal(amp, y[0], err_msg = "pulsed exp amp")
    
    # test non-relaxing
    y = pexp(x, 0, amp)
    assert_almost_equal(y[-1], amp, err_msg = "pulsed exp non-relaxing behaviour")
    
    # test inf relaxing
    y = pexp(x, np.inf, amp)
    assert_almost_equal(y[0], 0, err_msg = "pulsed exp infinitly fast relaxing behaviour")
    
    # test beam off position
    x = np.linspace(pulse_len-1e-6, pulse_len+1e-6, 10000)
    y = pexp(x, 1, amp)
    
    ddy = np.diff(np.diff(y))
    idx = ddy == min(ddy)
    
    assert_almost_equal((x[:-2])[idx], pulse_len, err_msg = "pulsed exp beam off position")
    
def test_pulsed_strexp():
    
    # settings
    amp = 1
    tau = 1
    pulse_len = 4
    
    x = np.linspace(1e-9, 10, 10000)
    psexp = pulsed_strexp(lifetime = tau, pulse_len = pulse_len)
    pexp = pulsed_exp(lifetime = tau, pulse_len = pulse_len)
    
    # test equivalence with pexp
    y1 = pexp(x, 1, amp)
    y2 = psexp(x, 1, 1, amp)
    # 1e-6 is the error in the integration
    assert_array_almost_equal(y1, y2, decimal=6, err_msg = 'pulsed str exp comparison to pulsed single exp when beta = 1')
    
    # test amp
    x = np.linspace(1e-15, 1, 10)
    y = psexp(x, 1, 0.5, amp)
    assert_almost_equal(amp, y[0], decimal=3, err_msg = 'pulsed str exp amp')
    
    # test non-relaxing
    y = psexp(x, 0, 0.5, amp)
    assert_almost_equal(y[0], y[-1], decimal=3, err_msg = 'pulsed str exp non-relaxing behaviour')
    
    # test inf relaxing
    y = psexp(x, np.inf, 0.5, amp)
    assert_almost_equal(y[0], 0, err_msg = 'pulsed str exp infinitly fast relaxing behaviour')
    
    # test beam off position
    x = np.linspace(pulse_len-1e-6, pulse_len+1e-6, 10000)
    y = psexp(x, 1, 0.5, amp)
    
    ddy = np.diff(np.diff(y))
    idx = ddy == min(ddy)
    
    assert_almost_equal((x[:-2])[idx], pulse_len, err_msg = 'pulsed str exp beam off position')



# Terminal access
# Derek Fujimoto
# July 2018

from tkinter import *
from tkinter import ttk
from bfit import logger_name
import logging, os, sys
import matplotlib.pyplot as plt
import numpy as np

# ========================================================================== #
class popup_terminal(object):
    """
        Popup window for python interpreter access. 
    """

    # ====================================================================== #
    def __init__(self, bfit):
        self.bfit = bfit
        
        # get logger
        self.logger = logging.getLogger(logger_name)
        self.logger.info('Initializing')
        
        # make a new window
        self.win = Toplevel(bfit.mainframe)
        self.win.title('Run Python Commands')
        frame = ttk.Frame(self.win, relief='sunken', pad=5)
        
        # icon
        self.bfit.set_icon(self.win)
        
        # Key bindings
        self.win.bind('<Control-Key-Return>', self.do_run)             
        self.win.bind('<Control-Key-KP_Enter>', self.do_run)
        
        # text input 
        self.text = Text(frame, width=80, height=20, state='normal')
        instructions = ttk.Label(frame, text="Press ctrl+enter to execute "+\
                                 "text at and below cursor position.")
        
        # scrollbar
        scrollb = Scrollbar(frame, command=self.text.yview)
        self.text['yscrollcommand'] = scrollb.set
        
        # gridding
        frame.grid(row=0, column=0)
        instructions.grid(column=0, row=0, sticky=(E, W), pady=5)
        self.text.grid(row=1, column=0)
        scrollb.grid(row=1, column=1, sticky='nsew')
        
        # grid frame
        self.logger.debug('Initialization success. Starting mainloop.')
    
    # ====================================================================== #
    def do_run(self, event):
        """
            Run python commands
        """
        # toplevel access
        bfit = self.bfit
        
        # remove newline
        if event.keysym == 'Return':
            self.text.delete('insert-1c')
        
        # get full text at and after cursor
        line_num = '%d.0' % int(float(self.text.index(INSERT)))
        lines = self.text.get(line_num, END)
        
        # run commands
        self.logger.info('Commands run: "%s"'% ('", "'.join(lines.split('\n')[:-1])))
        
        exec(lines)
        
    # ====================================================================== #
    def cancel(self):
        self.win.destroy()



# Put a function on the active plot and adjust it's parameters with mouse buttons
# Derek Fujimoto
# April 2019

import bdata as bd
import matplotlib.pyplot as plt
import matplotlib.patches as patches
import numpy as np
from bfit.fitting.functions import lorentzian # freq, peak, fwhm, amp
from bfit.fitting.functions import gaussian # freq, peak, fwhm, amp
from bfit.fitting.functions import bilorentzian # freq, peak, fwhm, amp
from bfit.fitting.functions import quadlorentzian # freq, nu_0, nu_q, eta, theta, 
                                                  # phi, amp0, amp1, amp2, amp3, 
                                                  #  fwhm0, fwhm1, fwhm2, fwhm3, I
from bfit.fitting.functions import pulsed_exp # time, lambda_s, amp
from bfit.fitting.functions import pulsed_strexp # time, lambda_s, beta, amp
from bfit.fitting.functions import qp_nu # nu_0, nu_q, eta, theta, phi, I, m

from functools import partial

class FunctionPlacer(object):
    
    npts = 1000  # number of points used to draw line
    
    # ======================================================================= #
    def __init__(self, fig, data, fn_single, ncomp, p0, fnname, endfn, asym_mode, 
                base=0, spin=2):
        """
            fig:    pointer to matplotlib figure object to draw in
            data:   bdata object 
            fn:     function to draw and get the parameters for
            p0:     dictionary of StringVar corresponding to input parameters
            endfn:  function pointer to function to call at end of sequence. 
                        Called with no inputs
            asym_mode: asymmetry calcuation type (c, sl_c, or dif_c)
            base:   value of the baseline when we're not altering it
            spin:   nuclear spin of probe I
        
            fn needs input parameters with keys: 
            
                1F/2E/1W
                    peak, fwhm, amp, base OR 
                    
                20/2H
                    amp, lam, beta (optional)
        """
        # save input
        self.fig = fig
        self.ncomp = ncomp
        self.base = base
        self.fn = lambda x, **kwargs : fn_single(x, **kwargs)+self.base
        self.fname = fnname
        self.endfn = endfn
        x = data.asym(asym_mode)[0]
        self.x = np.linspace(min(x), max(x), self.npts)
        self.spin = spin
        
        # get axes for drawing
        self.ax = fig.axes
        if len(self.ax) == 0:
            self.ax = fig.add_subplot(111)
        else:
            self.ax = self.ax[0]
        
        # get ylims
        ylims = self.ax.get_ylim()
    
        # make list of initial paramters 
        self.p0 = [{k:float(p[k].get()) for k in p.keys() if 'base' not in k} for p in p0]
    
        # baseline 
        if self.fname in ('Lorentzian', 'Gaussian', 'BiLorentzian', 'QuadLorentz'):
            self.base = float(p0[0]['base'].get())
            y = np.ones(len(self.x))*self.base
            self.baseline = self.ax.plot(self.x, y, zorder=20, ls='--')[0]
        else:
            self.base = 0
            self.baseline = dummybaseline()
    
        # draw each line with initial parameters
        self.lines = [self.ax.plot(self.x, 
                                   fn_single(self.x, **self.p0[i])+self.base, 
                                   zorder=20, ls='--')[0] for i in range(ncomp)]
        
        # make and draw function for the sum 
        self.sumfn = lambda x: np.sum([fn_single(x, **p) for p in self.p0], axis=0)+self.base
        self.sumline = self.ax.plot(self.x, self.sumfn(self.x), zorder=21, ls='-')[0]
        
        # reset ylims
        self.ax.set_ylim(ylims)
        
        # make title 
        self.ax.set_title('Press enter to save parameters', fontsize='small')
        self.fig.tight_layout()
        
        # connect enter key
        self.fig.canvas.mpl_connect('key_release_event', self.do_end)
        
        # resonance measurements ----------------------------------------------
        if self.fname in ('Lorentzian', 'Gaussian'):
            
            # if points are not saved they are garbage collected
            self.list_points = {'peak':[], 'fwhm':[]}
            
            # make points
            for i, (p, line) in enumerate(zip(self.p0, self.lines)):
                peakpt, widthpt = self.run_1f_single(p, line, 'C%d'%(i+1))
                self.list_points['peak'].append(peakpt)
                self.list_points['fwhm'].append(widthpt)
        
            self.list_points['base'] = self.run_1f_base(self.list_points['fwhm'], 'C0')
        
        elif self.fname == 'BiLorentzian':
            
            # if points are not saved they are garbage collected
            self.list_points = {'peak':[], 'fwhm':[]}
            
            # make points
            for i, (p, line) in enumerate(zip(self.p0, self.lines)):
                peakpt, widthpt = self.run_1f_bi_single(p, line, 'C%d'%(i+1))
                self.list_points['peak'].append(peakpt)
                self.list_points['fwhm'].append(widthpt)
        
            self.list_points['base'] = self.run_1f_base(self.list_points['fwhm'], 'C0')
        
        elif self.fname == 'QuadLorentz':
            
            # if points are not saved they are garbage collected
            self.list_points = {'peak0':[], 'peak1':[], 'peak2':[], 'peak3':[], 'fwhm':[], }
            
            # make points
            for i, (p, line) in enumerate(zip(self.p0, self.lines)):
                ppt0, ppt1, ppt2, ppt3, widthpt = self.run_1f_quad_single(p, line, 'C%d'%(i+1))
                self.list_points['peak0'].append(ppt0)
                self.list_points['peak1'].append(ppt1)
                self.list_points['peak2'].append(ppt2)
                self.list_points['peak3'].append(ppt3)
                self.list_points['fwhm'].append(widthpt)
            self.list_points['base'] = self.run_1f_quad_base(self.p0[0], self.list_points['fwhm'], 'C0')
            
        # SLR measurements ----------------------------------------------------
        elif self.fname in ('Exp', 'Str Exp'):
            
            # if points are not saved they are garbage collected
            self.list_points = {'amp':[], 'lam':[]}
            
            if self.fname == 'Str Exp':
                self.list_points['beta'] = []
            
            # make points
            for i, (p, line) in enumerate(zip(self.p0, self.lines)):
                self.list_points['amp'].append(self.run_20_initial(p, line, 'C%d'%i))
                self.list_points['lam'].append(self.run_20_lambda(p, line, 'C%d'%i))
                
                if self.fname == 'Str Exp':
                    self.list_points['beta'].append(self.run_20_beta(p, line, 'C%d'%i))
            
            # connect point shifter
            self.fig.canvas.mpl_connect('button_release_event', self.shift_20_pts_resize)
            
    # ======================================================================= #
    def run_1f_base(self, widths, color):
        """
            widths: list of points for widths, need to update y values
        """
        
        def update_base(x, y):
            
            # base point
            oldbase = self.base
            self.base = y
            for i in range(len(self.p0)):
                self.p0[i]['amp'] -= oldbase-y
            
            # update width points
            for p0, wpoint, line in zip(self.p0, widths, self.lines): 
                wpoint.point.set_ydata(self.fn(p0['peak']+p0['fwhm']/2, **p0))
                
                # update other lines
                line.set_ydata(self.fn(self.x, **p0))
            
            # update sumline
            self.sumline.set_ydata(self.sumfn(self.x))
            
            # update baseline line
            self.baseline.set_ydata(np.ones(self.npts)*self.base)    
            self.fig.canvas.draw()    
             
        # return so matplotlib doesn't garbage collect
        xpt = self.ax.get_xticks()[-1]
        return DraggablePoint(self, update_base, xpt, 
                                 self.base, 
                                 color=color, setx=False)
    
    # ======================================================================= #
    def run_1f_single(self, p0, line, color):
        
        # make point for width
        def update_width(x, y):
            
            # width point
            p0['fwhm'] = abs(p0['peak']-x)*2
            
            # update line
            line.set_ydata(self.fn(self.x, **p0))
            
            # update sum line
            self.sumline.set_ydata(self.sumfn(self.x))
            self.fig.canvas.draw()    
             
        x = p0['peak']+p0['fwhm']/2
        widthpt = DraggablePoint(self, update_width, x, self.fn(x, **p0), 
                                 color=color, sety=False)
        
        # make point for peak
        def update_peak(x, y):
        
            # peak point
            p0['amp'] = self.base-y
            p0['peak'] = x
        
            # width point
            x2 = x+p0['fwhm']/2
            widthpt.point.set_xdata(x2)
            widthpt.point.set_ydata(self.fn(x2, **p0))
            
            # update line
            line.set_ydata(self.fn(self.x, **p0))
            
            # update sum line
            self.sumline.set_ydata(self.sumfn(self.x))
            self.fig.canvas.draw()    
        
        peakpt = DraggablePoint(self, update_peak, p0['peak'], 
                                self.base-p0['amp'], color=color)
        
        return (peakpt, widthpt)
    
    # ======================================================================= #
    def run_1f_bi_single(self, p0, line, color):
        
        # make point for width
        def update_width(x, y):
            
            # width point
            p0['fwhm'] = abs(p0['peak']-x)*2
            
            # update line
            line.set_ydata(self.fn(self.x, **p0))
            
            # update sum line
            self.sumline.set_ydata(self.sumfn(self.x))
            self.fig.canvas.draw()    
             
        x = p0['peak']+p0['fwhm']/2
        widthpt = DraggablePoint(self, update_width, x, self.fn(x, **p0), 
                                 color=color, sety=False)
        
        # make point for peak
        def update_peak(x, y):
        
            # peak point
            p0['amp'] = self.base-y
            p0['peak'] = x
        
            # width point
            x2 = x+p0['fwhm']/2
            widthpt.point.set_xdata(x2)
            widthpt.point.set_ydata(self.fn(x2, **p0))
            
            # update line
            line.set_ydata(self.fn(self.x, **p0))
            
            # update sum line
            self.sumline.set_ydata(self.sumfn(self.x))
            self.fig.canvas.draw()    
        
        peakpt = DraggablePoint(self, update_peak, p0['peak'], 
                                self.base-p0['amp'], color=color)
        
        return (peakpt, widthpt)
     
    # ======================================================================= #
    def run_1f_quad_base(self, p0, widths, color):
        """
            widths: list of points for widths, need to update y values
        """
        peak0 = qp_nu(p0['nu_0'], p0['nu_q'], p0['eta'], p0['theta'], p0['phi'], \
                          self.spin, -1)
        
        def update_base(x, y):
            
            # base point
            oldbase = self.base
            self.base = y
        
            for i in range(len(self.p0)):
                self.p0[i]['amp0'] -= oldbase-y
                self.p0[i]['amp1'] -= oldbase-y
                self.p0[i]['amp2'] -= oldbase-y
                self.p0[i]['amp3'] -= oldbase-y
            
            # update width points
            for p0, wpoint, line in zip(self.p0, widths, self.lines): 
                wpoint.point.set_ydata(self.fn(peak0+p0['fwhm']/2, **p0))
                
                # update other lines
                line.set_ydata(self.fn(self.x, **p0))
            
            # update sumline
            self.sumline.set_ydata(self.sumfn(self.x))
            
            # update baseline line
            self.baseline.set_ydata(np.ones(self.npts)*self.base)    
            self.fig.canvas.draw()    
             
        # return so matplotlib doesn't garbage collect
        xpt = self.ax.get_xticks()[-1]
        
        return DraggablePoint(self, update_base, xpt, 
                                 self.base, 
                                 color=color, setx=False)
        
    # ======================================================================= #
    def run_1f_quad_single(self, p0, line, color):
        """
            p0 keys: 'amp0', 'amp1', 'amp2', 'amp3', 'eta', 'phi', 'theta', 'fwhm', 'nu_0', 'nu_q'
        """
        
        # lorentzian fn: lorentzian # freq, peak, width, amp
        s = self.spin
        
        # peak locations from right to left
        peak = lambda i: qp_nu(p0['nu_0'], p0['nu_q'], p0['eta'], p0['theta'], \
                               p0['phi'], s, i-s+1)
        
        # set width
        x = peak(0)+p0['fwhm']/2
        widthpt = DraggablePoint(self, None, x, self.fn(x, **p0), 
                                 color=color, sety=False)
        
        # set peak and amplitudes
        peakpts = []
        for n in range(2*s):
            x = p0['nu_0'] + peak(n) - (peak(s+1-n)+peak(n))/2
            
            y = self.base - \
                p0['amp%d'%n] + \
                sum([lorentzian(peak(n), 
                                peak(i%(2*s)), 
                                p0['fwhm'], 
                                p0['amp%d'%(i%(2*s))]) for i in range(n+1, 2*s+n)])                    
            
            if n in (0, 2*s-1):
                peakpts.append(DraggablePoint(self, None, x, y, color=color, marker='s'))
            else:
                peakpts.append(DraggablePoint(self, None, x, y, color=color, marker='^', setx=False))
            
        # make point for width
        def update_width(x, y):
                        
            # width point
            p0['fwhm'] = abs(peak(0)-x)*2
            
            # update line
            line.set_ydata(self.fn(self.x, **p0))
            
            # update peak heights
            for i in range(2*s):
                peakpts[i].point.set_ydata(self.fn(peak(i), **p0))
            
            # update width y 
            widthpt.point.set_ydata(self.fn(x, **p0))
            
            # update sum line
            self.sumline.set_ydata(self.sumfn(self.x))
            self.fig.canvas.draw()    

        # make point for peak, updating nu_0
        def update_peak_center(x, y, n):
            
            # peak point
            p0['amp%d'%n] = self.base - y + \
                        sum([lorentzian(peak(n), 
                                        peak(i%(2*s)), 
                                        p0['fwhm'], 
                                        p0['amp%d'%(i%(2*s))]) for i in range(n+1, 2*s+n)])
            
            # width point        
            x2 = peak(0)+p0['fwhm']/2
            widthpt.point.set_ydata(self.fn(x2, **p0))
        
            # update the other peak points
            for i in range(n+1, 2*s+n):
                peakpts[i%(2*s)].point.set_ydata(self.fn(peak(i%(2*s)), **p0))
            
            # update line
            line.set_ydata(self.fn(self.x, **p0))
            
            # update sum line
            self.sumline.set_ydata(self.sumfn(self.x))
            self.fig.canvas.draw()    
        
        # make point for peak, updating nu_q 
        def update_peak_edge(x, y, n):
            
            # amplitude
            p0['amp%d'%n] = self.base - y + \
                        sum([lorentzian(peak(n), 
                                        peak(i%(2*s)), 
                                        p0['fwhm'], 
                                        p0['amp%d'%(i%(2*s))]) for i in range(n+1, 2*s+n)])
            
            # get x and n of the other edge peak position
            other_n = 2*s-n-1
            other_x = float(peakpts[other_n].point.get_xdata())
            
            # set nu_0
            p0['nu_0'] = (other_x+x)/2
            
            # set nu_q
            # Equation (28)
            V_0 = np.sqrt(1.5) * 0.5 * (3 * np.square(np.cos(p0['theta'])) \
                        - 1 + p0['eta'] * np.square(np.sin(p0['theta'])) * \
                        np.cos(2 * p0['phi']))
                        
            # Equation (23)
            p0['nu_q'] = (x-p0['nu_0']) / ((np.sqrt(6) / 3) * (1 - 2 * (n-s+1)) * V_0)  
            
            # width point        
            x2 = peak(0)+p0['fwhm']/2
            widthpt.point.set_xdata(x2)
            widthpt.point.set_ydata(self.fn(x2, **p0))
        
            # update the other peak points
            for i in range(1, 2*s-1):
                peakpts[i].point.set_xdata(peak(i))
                peakpts[i].point.set_ydata(self.fn(peak(i), **p0))
            peakpts[other_n].point.set_ydata(self.fn(other_x, **p0))
            
            # update line
            line.set_ydata(self.fn(self.x, **p0))
            
            # update sum line
            self.sumline.set_ydata(self.sumfn(self.x))
            self.fig.canvas.draw()    
        
        widthpt.updatefn = update_width
        peakpts[0].updatefn = partial(update_peak_edge, n=0)
        peakpts[2*s-1].updatefn = partial(update_peak_edge, n=2*s-1)
        for i in range(1, 2*s-1):
            peakpts[i].updatefn = partial(update_peak_center, n=i)
        
        return (*peakpts, widthpt)
        
    # ======================================================================= #
    def run_20_initial(self, p0, line, color):
        def update(x, y):
            # initial asymmetry
            p0['amp'] = max(0, y)
            line.set_ydata(self.fn(self.x, **p0))
            
            # update sumline
            self.sumline.set_ydata(self.sumfn(self.x))
            self.fig.canvas.draw()    
             
        # return so matplotlib doesn't garbage collect
        return DraggablePoint(self, update, 0, p0['amp'], 
                              color=color, setx=False)
    
    # ======================================================================= #
    def run_20_lambda(self, p0, line, color):
        def update(x, y):
            # 1/T1
            p0['lam'] = max(0, 1/x)
            line.set_ydata(self.fn(self.x, **p0))
            
            # update sumline
            self.sumline.set_ydata(self.sumfn(self.x))
            self.fig.canvas.draw()    
             
        # return so matplotlib doesn't garbage collect
        x = np.ones(1)/p0['lam']
        ylim = self.ax.get_ylim()[0]
        y = min([i for i in self.ax.get_yticks() if i>ylim])
        return DraggablePoint(self, update, x, y, 
                              color=color, sety=False)
    
    # ======================================================================= #
    def run_20_beta(self, p0, line, color):
        def update(x, y):
    
            # beta - put y axis on a range of 0 to 1
            ylo, yhi = self.ax.get_ylim()
            y = (y-ylo)/(yhi-ylo)
            p0['beta'] = y
            line.set_ydata(self.fn(self.x, **p0))
            
            # update sumline
            self.sumline.set_ydata(self.sumfn(self.x))
            self.fig.canvas.draw()    
             
        # return so matplotlib doesn't garbage collect
        ylo, yhi = self.ax.get_ylim()
        xlo, xhi = self.ax.get_xlim()
        x = max([i for i in self.ax.get_xticks() if i < xhi])
        return DraggablePoint(self, update, x, p0['beta']*(yhi-ylo)+ylo, 
                              color=color, setx=False)
    
    # ======================================================================= #
    def shift_20_pts_resize(self, event):
        """Move the points so they fit in the screen"""
        
        xlo, xhi = self.ax.get_xlim()
        ylo, yhi = self.ax.get_ylim()
        
        for lam in self.list_points['lam']:
            ypos = lam.point.get_ydata()
            lam.point.set_ydata(min([i for i in self.ax.get_yticks() if i>ylo]))
        
        if 'beta' in self.list_points.keys():
            for beta in self.list_points['beta']:
                xpos = beta.point.get_xdata()
                beta.point.set_xdata(max([i for i in self.ax.get_xticks() if i<xhi])) 
    
    # ======================================================================= #
    def do_end(self, event):
        if event.key == 'enter':
            plt.close(self.fig.number)
            self.endfn(self.p0, self.base)
        
class dummybaseline(object):
    def __init__(self):pass
    def set_ydata(self, y) : pass
                
class DraggablePoint:

    # http://stackoverflow.com/questions/21654008/matplotlib-drag-overlapping-points-interactively
    # https://stackoverflow.com/questions/28001655/draggable-line-with-draggable-points
    
    lock = None #  only one can be animated at a time
    size=0.01

    # ======================================================================= #
    def __init__(self, parent, updatefn, x, y, setx=True, sety=True, color=None, marker='s'):
        """
            parent: parent object
            updatefn: funtion which updates the line in the correct way
                updatefn(xdata, ydata)
            x, y: initial point position
            setx, sety: if true, allow setting this parameter
            color: point color
        """
        self.parent = parent
        self.point = parent.ax.plot(x, y, zorder=25, color=color, alpha=0.5, 
                                 marker=marker, markersize=8)[0]
        
        self.point.set_pickradius(8)
        self.updatefn = updatefn
        self.setx = setx
        self.sety = sety
        self.press = None
        self.background = None
        self.connect()
        
    # ======================================================================= #
    def connect(self):
        """connect to all the events we need"""
        self.cidpress = self.point.figure.canvas.mpl_connect('button_press_event', self.on_press)
        self.cidrelease = self.point.figure.canvas.mpl_connect('button_release_event', self.on_release)
        self.cidmotion = self.point.figure.canvas.mpl_connect('motion_notify_event', self.on_motion)

    # ======================================================================= #
    def on_press(self, event):
        if event.inaxes != self.point.axes: return
        if DraggablePoint.lock is not None: return
        contains, attrd = self.point.contains(event)
        if not contains: return
        DraggablePoint.lock = self
        
    # ======================================================================= #
    def on_motion(self, event):

        if DraggablePoint.lock is not self: return
        if event.inaxes != self.point.axes: return
        
        # get data
        x = event.xdata
        y = event.ydata
        
        # move the point
        if self.setx:   self.point.set_xdata(x)
        if self.sety:   self.point.set_ydata(y)

        # update the line
        self.updatefn(x, y)        

    # ======================================================================= #
    def on_release(self, event):
        'on release we reset the press data'
        if DraggablePoint.lock is not self: return
        DraggablePoint.lock = None
        
    # ======================================================================= #
    def disconnect(self):

        'disconnect all the stored connection ids'

        self.point.figure.canvas.mpl_disconnect(self.cidpress)
        self.point.figure.canvas.mpl_disconnect(self.cidrelease)
        self.point.figure.canvas.mpl_disconnect(self.cidmotion)



# test fit results with model
from numpy.testing import *
import numpy as np
from bfit.gui.popup_fit_results import popup_fit_results
from bfit.gui.bfit import bfit

# filter unneeded warnings
import pytest
pytestmark = pytest.mark.filterwarnings('ignore:2020')

def with_bfit(function):
    
    def wrapper(*args, **kwargs):
        # make gui
        b = bfit(None, True)
        tab = b.fit_files
        tab2 = b.fetch_files
        b.notebook.select(2)
        
        tab2.year.set(2020)
        tab2.run.set('40123 40127, 40129')
        tab2.get_data()
        tab.populate()
        b.draw_fit.set(False)
        
        try:
            return function(*args, **kwargs, b=b, fittab=tab, fetchtab=tab2)
        finally:
            b.on_closing()
            del b
            
    return wrapper

@with_bfit
def test_par_detection(b=None, fittab=None, fetchtab=None):
    
    # set inputs
    res = popup_fit_results(b)
    res.entry.insert('1.0', 'y = a*x+b')
    res.get_input()
    
    assert all([k in res.new_par['name'].values for k in 'ab'])
    assert len(res.new_par['name'].values) == 2

    
@with_bfit
def test_fit_accuracy(b=None, fittab=None, fetchtab=None):
    
    # set inputs
    res = popup_fit_results(b)
    res.entry.insert('1.0', 'y = a*x+b')
    res.get_input()
    
    res.xaxis.set('Temperature (K)')
    res.yaxis.set('Temperature (K)')
    
    # do fit
    res.do_fit()
    
    # check chi2
    assert res.chi < 0.1



# test minuit object
# Derek Fujimoto
# Feb 2021

from numpy.testing import *
from bfit.fitting.minuit import minuit
import numpy as np

# test inputs
fn = lambda x, a, b, c, d, e: a+b+c+d+e+x
fn2 = lambda x, par: a+b+c+d+e+x
x = np.arange(10)
y = x**2

def test_start():
    m = minuit(fn, x, y, start=[1,2,3,4,5])
    assert_equal(m.values['e'], 5, 'minuit start assignment named')
    
    m = minuit(fn2, x, y, start=[1,2,3,4,5])
    assert_equal(m.values['x4'], 5, 'minuit start assignment unnamed')
    
    m = minuit(fn, x, y, e=5)
    assert_equal(m.values['e'], 5, 'minuit start assignment single named')
    
    m = minuit(fn, x, y)
    assert_equal(m.values['e'], 1, 'minuit start default param named')
    
def test_name_assign():
    m = minuit(fn2, x, y, name=['a', 'b', 'c', 'd', 'e'])
    assert m.values['e'], 'minuit name assignment'
    
def test_name_default():
    m = minuit(fn2, x, y, name=['a', 'b', 'c', 'd', 'e'])    
    assert_equal(m.values['e'], 1, 'minuit start default param unnamed')
    
def test_error_assign():
    m = minuit(fn, x, y, error=[1,2,3,4,5])
    assert_equal(m.errors['e'], 5, 'minuit errors assignment')

def test_error_broadcast():    
    m = minuit(fn, x, y, error=5)
    errors_all = [m.errors[k] == 5 for k in 'abcde']
    assert_equal(all(errors_all), True, 'minuit errors broadcasting')
    
def test_error_named_assign():
    m = minuit(fn, x, y, error_e=5)
    assert_equal(m.errors['e'], 5, 'minuit errors named assignment')
    
def test_error_named_and_list_assign():
    m = minuit(fn, x, y, error_e=5, error=1)
    assert_equal(m.errors['e'], 5, 'minuit errors list and named assignment')
    
    others = [m.errors[k]==1 for k in 'abcd']
    assert_equal(all(others), True, 'minuit errors list and named assignment')
    
def test_limit_assign():
    m = minuit(fn, x, y, limit=[[0,1],[0,2],[0,3],[0,4],[0,5]])
    assert_equal(m.limits['e'][1], 5, 'minuit limits assignment')    
    
def test_limit_broadcast():
    m = minuit(fn, x, y, limit=[0,5])
    limits_all = [m.limits[k][1] == 5 for k in 'abcde']
    assert_equal(all(limits_all), True, 'minuit limits broadcasting')    
    
    limits_all = [m.limits[k][0] == 0 for k in 'abcde']
    assert_equal(all(limits_all), True, 'minuit limits broadcasting')    
    
def test_limit_named_assign():
    m = minuit(fn, x, y, limit_e=[0,5])
    assert_equal(m.limits['e'][0], 0, 'minuit limits 0 named assignment')
    assert_equal(m.limits['e'][1], 5, 'minuit limits 1 named assignment')

def test_limit_named_and_list_assign():
    m = minuit(fn, x, y, limit_e=[1,2], limit=[0,5])
    
    assert_equal(m.limits['e'][0], 1, 'minuit limits 0 list and named assignment')
    assert_equal(m.limits['e'][1], 2, 'minuit limits 1 list and named assignment')
    
    others = [m.limits[k][0]==0 for k in 'abcd']
    assert_equal(all(others), True, 'minuit limits 0 list and named assignment')
    
    others = [m.limits[k][1]==5 for k in 'abcd']
    assert_equal(all(others), True, 'minuit limits 1 list and named assignment')
    
def test_fix_assign():
    m = minuit(fn, x, y, fix=[True, True, True, True, True])
    assert_equal(m.fixed['e'], True, 'minuit fixed assignment')
    
def test_fix_broadcast():
    m = minuit(fn, x, y, fix=True)
    assert_equal(m.fixed['e'], True, 'minuit fixed broadcasting')
    
def test_fix_named_assign():
    m = minuit(fn, x, y, fix_e=True)
    assert_equal(m.fixed['e'], True, 'minuit fixed named assignment')
    
    fixed_other = [m.fixed[k] for k in 'abcd']
    assert_equal(any(fixed_other), False, 'minuit fixed named assignment, other')
    
def test_fix_named_and_list_assign():
    m = minuit(fn, x, y, fix_e=False, fix=True)
    assert_equal(m.fixed['e'], False, 'minuit fixed list and named assignment')
    
    others = [m.fixed[k] for k in 'abcd']
    assert_equal(all(others), True, 'minuit fixed list and named assignment')



from .fitting.functions import lorentzian, bilorentzian, gaussian, quadlorentzian
from .fitting.functions import pulsed_exp, pulsed_strexp, pulsed_biexp
from .fitting.global_fitter import global_fitter
from .fitting.global_bdata_fitter import global_bdata_fitter
from .fitting.fit_bdata import fit_bdata
from .fitting.minuit import minuit

import logging, os, sys, argparse, subprocess, requests, json, code
from logging.handlers import RotatingFileHandler
from textwrap import dedent
from pkg_resources import parse_version 
from tkinter import messagebox

__all__ = ['gui','fitting','backend','test']
__version__ = '4.9.7'
__author__ = 'Derek Fujimoto'
logger_name = 'bfit'
icon_path = os.path.join(os.path.dirname(__file__),'data','icon.gif')

__all__.extend(("lorentzian", 
                "bilorentzian", 
                "gaussian", 
                "quadlorentzian",
                "pulsed_exp",
                "pulsed_strexp",
                "pulsed_biexp",
                "global_fitter",
                "global_bdata_fitter",
                "fit_bdata",
                "minuit",
                ))

from bfit.gui.bfit import bfit

# RUN BFIT ================================================================== #
def main():
    
    # check if maxOS
    if sys.platform == 'darwin':
        os.environ['OBJC_DISABLE_INITIALIZE_FORK_SAFETY'] = 'YES'

    # command line switches ---------------------------------------------------
    parser = argparse.ArgumentParser(description=dedent("""\
        Run BNMR data viewer and fitter for online application."""), 
        formatter_class=argparse.RawTextHelpFormatter)
    
    # logging level
    parser.add_argument("-d", "--debug", 
                        help='Run in debug mode', 
                        dest='debug', 
                        action='store_true', 
                        default=False)
    
    # no gui mode
    parser.add_argument("-c", "--commandline", 
                        help='Run in command line mode', 
                        dest='commandline', 
                        action='store_true', 
                        default=False)

    # parse
    args = parser.parse_args()

    # Setup logging -----------------------------------------------------------
    logger = logging.getLogger(logger_name)

    # get log filename
    try:
        filename = os.path.join(os.environ['HOME'], '.bfit.log')
    except KeyError:
        filename = 'bfit.log'

    # make handler
    handler = RotatingFileHandler(filename, 
                                  mode='a', 
                                  maxBytes=100*1000, # 100 kB max
                                  backupCount=1)

    # get level and format for output string
    if args.debug:
        level = logging.DEBUG
        fmt = '%(asctime)s %(levelname)-8s %(module)s.%(funcName)s() [%(lineno)d] -- %(message)s'
    else:
        level = logging.INFO
        fmt = '%(asctime)s %(levelname)-8s %(module)s -- %(message)s'
    
    # set
    handler.setFormatter(logging.Formatter(fmt))
    handler.setLevel(level)

    logger.addHandler(handler)
    logger.setLevel(level)
    
    # testing
    testfn = None
    # ~ def testfn(self):
        # ~ self.fetch_files.run.set("40123, 40127")
        # ~ self.fetch_files.year.set(2012)
        # ~ self.fetch_files.get_data()
        # ~ self.fit_files.fit_function_title.set("QuadLorentz")
        # ~ self.fit_files.populate()
        # ~ self.notebook.select(2)
        # ~ self.fit_files.do_fit()
        # ~ import matplotlib.pyplot as plt
        # ~ plt.close('all')
        # ~ self.fit_files.do_add_param()
        
    # Check version (credit: https://github.com/alexmojaki/outdated) ----------
    try:
        latest_version = requests.get('https://pypi.python.org/pypi/bfit/json').text
        latest_version = json.loads(latest_version)['info']['version']
        latest_version2 = parse_version(latest_version)
        current_version = parse_version(str(__version__))
        if current_version < latest_version2: 
            testfn = lambda x: messagebox.showinfo("Please update", 
                                "Version %s available!" % latest_version)
    except Exception:
        pass
    
    # start bfit --------------------------------------------------------------
    b = bfit(testfn, args.commandline)
    if args.commandline:
        code.interact(local=locals())
        print('bfit object set to variable "b"')
    

    
        



# test inspect tab
# Derek Fujimoto
# Feb 2021

from numpy.testing import *
import numpy as np
from bfit.gui.calculator_nmr_atten import calculator_nmr_atten

def test_calc_power():
    calc = calculator_nmr_atten(True)
    
    # power
    calc.power.set('100')
    calc.entry_power.focus_set()
    calc.calculate()
    assert_almost_equal(float(calc.dac.get()), 0, 
                        err_msg = "calculator nmr atten set power = 100%")
    
    calc.power.set('0')
    calc.entry_power.focus_set()
    calc.calculate()
    assert_almost_equal(float(calc.dac.get()), 2047, 
                        err_msg = "calculator nmr atten set power = 0%")
    calc.root.destroy()
    
def test_calc_dac():
    calc = calculator_nmr_atten(True)
    
    # dac
    calc.dac.set('0')
    calc.entry_dac.focus_set()
    calc.calculate()
    assert_almost_equal(float(calc.power.get()), 100, 
                        err_msg = "calculator nmr atten set dac = 0")
    
    calc.dac.set('2047')
    calc.entry_dac.focus_set()
    calc.calculate()
    assert_almost_equal(float(calc.power.get()), 0, 
                        err_msg = "calculator nmr atten set dac = 2047")
    calc.root.destroy()
    



# test save and load state
from numpy.testing import *
import numpy as np
import pandas as pd
import bdata as bd
from bfit.gui.bfit import bfit
import os

filename = 'test.yaml'

menu     = (('style', 'alpha', 0.1),
            ('hist_select', 'B+,B-,F+,F-'),
            ('label_default', "1000/T (1/K)"),
            ('ppm_reference', 20),
            ('probe_species', 'F20'),
            ('update_period', 2),
            ('thermo_channel', 'B'),
            ('units', '1f', [1, 'TEST_HZ']),
            ('draw_style', 'new'),
            ('draw_fit', False),
            ('draw_ppm', True),
            ('draw_standardized_res', False),
            ('norm_with_param', False),
            ('use_nbm', True),
            ('minimizer', 'bfit.fitting.fitter_migrad_minos'),
            )
            
fileview = (('year', 2019),
            ('runn', 40123),
            ('is_updating', True),
            ('asym_type', 'Split Helicity'),
            ('rebin', 20),
            )
            
fetch    = (('year', 2019),
            ('run', '40123 40127'),
            ('check_state', False),
            ('check_state_data', False),
            ('check_state_fit', True),
            ('check_state_res', True),
            ('check_rebin', 2),
            ('check_bin_remove', '10-20'),
            ('asym_type', 'Shifted Split'),
            )
            
dataline = (('check_state', True),
            ('label', 'TEST LABEL'),
            ('rebin', 10),
            ('bin_remove', '10 22'),
            )
            
fit      = (('annotation', 'fwhm'),
            ('asym_type', 'Positive Helicity'),
            ('fit_function_title', 'Gaussian'),
            ('n_component', 2),
            ('par_label', 'testlabel'),
            ('set_as_group', True),
            ('set_prior_p0', True),
            ('use_rebin', True),
            ('xaxis', 'baseline'),
            ('yaxis', 'height'),
            ('xlo', '2.5'),
            ('xhi', '8.6'),
            )
            
fitline  = (('parentry', 'baseline', 'p0', 1),
            ('parentry', 'baseline', 'blo', 2),
            ('parentry', 'baseline', 'bhi', 3),
            ('parentry', 'baseline', 'res', 4),
            ('parentry', 'baseline', 'dres+', 5),
            ('parentry', 'baseline', 'dres-', 6),
            ('parentry', 'baseline', 'chi', 7),
            ('parentry', 'height_0', 'fixed', True),
            ('parentry', 'baseline', 'shared', True),
            )
            
deadtime = (('deadtime', 2),
            ('deadtime_switch', True),
            ('deadtime_global', False),
            )
            
datadir  = (('bnqr_data_dir', 'testdir_nqr'),
            )

def setv(obj_top, values):
    
    # get top level object
    obj = getattr(obj_top, values[0])
    
    # set value
    if hasattr(obj, 'set'):
        obj.set(values[1])
    
    # parentry dict
    elif values[0] == 'parentry':
        obj[values[1]][values[2]][0].set(values[3])
    
    # if dict, get dict value
    elif type(obj) is dict: 
        obj[values[1]] = values[2]
            
    # set value
    else:
        setattr(obj_top, values[0], values[1])

def getv(obj_top, values):
    
    # get top level object
    obj = getattr(obj_top, values[0])
    
    # parentry dict
    if values[0] == 'parentry':
        return (obj[values[1]][values[2]][0].get(), values[3])
    
    # if dict, get dict value
    elif type(obj) is dict: 
        return (obj[values[1]], values[2])

    # get value
    elif hasattr(obj, 'get'):
        return (obj.get(), values[1])
            
    # get value
    else:
        return (getattr(obj_top, values[0]), values[1])

def save():
    
    # make gui
    b = bfit(None, True)

    # set menu items -------------------------------------------------------
    for v in menu:
        setv(b, v)
    
    # set fileviewer items ---------------------------------------------------
    tab = b.fileviewer
    for i, v in enumerate(fileview):
        setv(tab, v)
        if i == 1:
            tab.get_data()
            
    # set fetch items --------------------------------------------------------
    tab = b.fetch_files
    for i, v in enumerate(fetch):
        setv(tab, v)
        if i == 1:
            tab.get_data()

    dline = tab.data_lines['2019.40123']
    for v in dataline:
        setv(dline, v)
    
    # set fit items ----------------------------------------------------------
    tab = b.fit_files
    tab.populate()
    tab.do_fit()
    for v in fit:
        setv(tab, v)
        
    tab.populate()
    
    fit_line = tab.fit_lines['2019.40123']
    for v in fitline:
        setv(fit_line, v)
    
    # misc -------------------------------------------------------------------
    
    # deadtime
    for v in deadtime:
        setv(b, v)
    
    # data directory
    for v in datadir:
        setv(b, v)
        
    # save and clean ---------------------------------------------------------
    b.fit_files.save_state(filename)
    b.on_closing()
    del b

def test_load():
    
    # save
    save()
    
    # make gui and load
    b = bfit(None, True)
    b.fit_files.load_state(filename)
    
    # check list
    def check(obj, lst):
        for v in lst:
            read, desired = getv(obj, v)
            try:
                assert read == desired, ('%s\n%s \n\tACTUAL: %s\n\tDESIRED: %s' % \
                    (str(obj), v[:-1], read, desired))
            except AssertionError as err:
                if type(read) != type(desired): 
                    assert str(read) == str(desired), ('%s\n%s \n\tACTUAL: %s\n\tDESIRED: %s' % \
                        (str(obj), v[:-1], read, desired))
                else:
                    raise err from None
    
    # check all the lists
    check(b, menu)
    check(b.fileviewer, fileview)
    check(b.fetch_files, fetch)
    check(b.fetch_files.data_lines['2019.40123'], dataline)
    check(b.fit_files, fit)
    check(b.fit_files.fit_lines['2019.40123'], fitline)
    check(b, deadtime)
    check(b, datadir)
    
    # save and clean ---------------------------------------------------------
    b.on_closing()    
    del b
    os.remove(filename)
    



# test exporting fit values
from numpy.testing import *
import numpy as np
import pandas as pd
import bdata as bd
from bfit.gui.bfit import bfit
import os

# filter unneeded warnings
import pytest
pytestmark = pytest.mark.filterwarnings('ignore:2020')

# test file name
filename = 'TESTFILE.csv'

def with_bfit(function):
    
    def wrapper(*args, **kwargs):
        # make gui
        b = bfit(None, True)
        tab = b.fetch_files
        tab2 = b.fit_files
        b.notebook.select(2)
        b.draw_fit.set(False)
        
        # get data
        tab.year.set(2020)
        tab.run.set('40123 40127')
        tab.get_data()
        
        # fit
        tab2.populate()
        tab2.do_fit()
        
        # test
        try:
            return function(*args, **kwargs, b=b)
        finally:
            b.on_closing()
            del b
            
    return wrapper

def check_columns(filename):
    
    # load file
    df = pd.read_csv(filename, comment='#')
    
    # check columns
    columns = ['Time (s)', 'asymmetry']
    for c in df.columns:
        assert c in columns
    
    # check column length
    assert len(df.columns) == len(columns)

def check_header(filename, data):
    
    # get header lines
    with open(filename, 'r') as fid:
        lines = fid.readlines()
    lines = [l[1:].strip() for l in lines if l[0] == '#']
    
    # strip labels
    lines = [l.split(':')[-1].strip() if i > 1 else l for i, l in enumerate(lines)]
    lines = [l for l in lines if l]
    
    # split lists
    for i in [6, 7, 8, 9]:
        lines[i] = lines[i].split(', ')
    
    # check id
    assert lines[0] == data.id
    
    # check title
    assert lines[1] == data.title
    
    # check fit function
    assert lines[2] == data.fit_title
    
    # check number components
    assert lines[3] == str(data.ncomp)
    
    # check rebin
    assert lines[4] == str(data.rebin.get())
    
    # check chisquared
    assert_almost_equal(float(lines[5]), data.chi, err_msg="chisquared header copy",
                        decimal=6)
    
    # check param names
    assert_array_equal(lines[6], data.parnames, err_msg='parameter names copy')
    
    # check par values, errors
    for i, name in enumerate(lines[6]):
        assert_almost_equal(float(lines[7][i]), data.fitpar.loc[name, 'res'], 
                            err_msg="Parameter value copy %s" % name)
    
        assert_almost_equal(float(lines[8][i]), data.fitpar.loc[name, 'dres-'], 
                            err_msg="Parameter error- copy %s" % name)
    
        assert_almost_equal(float(lines[9][i]), data.fitpar.loc[name, 'dres+'], 
                            err_msg="Parameter error+ copy %s" % name)

def check_content(filename, data):
        
    # get fit function
    fitfn = data.fitfn
    
    # read file
    df = pd.read_csv(filename, comment='#')
    
    x = df['Time (s)'].values
    y = df['asymmetry'].values
        
    # check content
    testy = fitfn(x, *data.fitpar['res'].values)
    assert_array_almost_equal(y, testy, err_msg='asymmetry content check %s' % filename)
    
@with_bfit
def test_columns(b=None):
    
    # export
    b.fit_files.export_fit(directory='.')
    
    # test
    check_columns('2020_40123_fit.csv')
    check_columns('2020_40127_fit.csv')
    
    # clean
    os.remove('2020_40123_fit.csv')
    os.remove('2020_40127_fit.csv')

@with_bfit
def test_header(b=None):
    
    # export
    b.fit_files.export_fit(directory='.')
    
    # test
    check_header('2020_40123_fit.csv', b.data['2020.40123'])
    check_header('2020_40127_fit.csv', b.data['2020.40127'])
    
    # clean
    os.remove('2020_40123_fit.csv')
    os.remove('2020_40127_fit.csv')
    
@with_bfit
def test_content(b=None):
    
    # export
    b.fit_files.export_fit(directory='.')
    
    # test
    check_content('2020_40123_fit.csv', b.data['2020.40123'])
    check_content('2020_40127_fit.csv', b.data['2020.40127'])
    
    # clean
    os.remove('2020_40123_fit.csv')
    os.remove('2020_40127_fit.csv')



# Test constrained fit module
from numpy.testing import *
import numpy as np
from bfit.gui.popup_fit_constraints import popup_fit_constraints
from bfit.gui.bfit import bfit

# filter unneeded warnings
import pytest
pytestmark = pytest.mark.filterwarnings('ignore:2020')

def with_bfit(function):
    
    def wrapper(*args, **kwargs):
        # make gui
        b = bfit(None, True)
        tab = b.fit_files
        tab2 = b.fetch_files
        b.notebook.select(2)
        
        tab2.year.set(2020)
        tab2.run.set('40123 40127')
        tab2.get_data()
        tab.populate()
        b.draw_fit.set(False)
        
        try:
            return function(*args, **kwargs, b=b, fittab=tab, fetchtab=tab2)
        finally:
            b.on_closing()
            del b
            
    return wrapper

@with_bfit
def test_par_detection(b=None, fittab=None, fetchtab=None):
    
    constr = popup_fit_constraints(b)
    constr.entry.insert('1.0', '1_T1 = a*np.exp(b*BIAS**0.5) + c\namp=d')
    constr.get_input()
    
    assert all([k in constr.new_par['name'].values for k in 'abcd'])
    assert len(constr.new_par['name'].values) == 4

@with_bfit
def test_defineable(b=None, fittab=None, fetchtab=None):
    
    constr = popup_fit_constraints(b)
    assert all([k in constr.parnames for k in ('1_T1', 'amp')])
    assert len(constr.parnames) == 2
    
@with_bfit
def test_fit_accuracy(b=None, fittab=None, fetchtab=None):
    
    constr = popup_fit_constraints(b)
    constr.entry.insert('1.0', '1_T1 = a\namp=b')
    constr.get_input()
    constr.do_fit()
    
    # get results
    out_c = constr.new_par.set_index('name')
    
    # fit with normal sharing
    line = fittab.fit_lines['2020.40123']
    entry = line.parentry
    entry['1_T1']['shared'][0].set(True)
    entry['amp']['shared'][0].set(True)
    fittab.do_fit()
    
    # check results
    assert_almost_equal(out_c.loc['a','res'], float(entry['1_T1']['res'][1].get()), 
                        err_msg = 'T1 res not equal', decimal=5)
    assert_almost_equal(out_c.loc['a','err+'], float(entry['1_T1']['dres+'][1].get()), 
                        err_msg = 'T1 err+ not equal', decimal=5)
    assert_almost_equal(out_c.loc['a','err-'], float(entry['1_T1']['dres-'][1].get()), 
                        err_msg = 'T1 err- not equal', decimal=5)
    
    assert_almost_equal(out_c.loc['b','res'], float(entry['amp']['res'][1].get()), 
                        err_msg = 'T1 res not equal', decimal=5)
    assert_almost_equal(out_c.loc['b','err+'], float(entry['amp']['dres+'][1].get()), 
                        err_msg = 'T1 err+ not equal', decimal=5)
    assert_almost_equal(out_c.loc['b','err-'], float(entry['amp']['dres-'][1].get()), 
                        err_msg = 'amp err- not equal', decimal=5)
    
@with_bfit
def test_fit_copy(b=None, fittab=None, fetchtab=None):
    
    constr = popup_fit_constraints(b)
    constr.entry.insert('1.0', '1_T1 = a\namp=b')
    constr.get_input()
    constr.do_fit()
    
    # get results
    out_c = constr.new_par.set_index('name')
    line = fittab.fit_lines['2020.40123']
    entry = line.parentry
    
    # check that results copied to main page properly
    assert_almost_equal(out_c.loc['a','res'], float(entry['1_T1']['res'][1].get()), 
                        err_msg = 'T1 res not equal', decimal=5)
    assert_almost_equal(out_c.loc['a','err+'], float(entry['1_T1']['dres+'][1].get()), 
                        err_msg = 'T1 err+ not equal', decimal=5)
    assert_almost_equal(out_c.loc['a','err-'], float(entry['1_T1']['dres-'][1].get()), 
                        err_msg = 'T1 err- not equal', decimal=5)
    
    assert_almost_equal(out_c.loc['b','res'], float(entry['amp']['res'][1].get()), 
                        err_msg = 'T1 res not equal', decimal=5)
    assert_almost_equal(out_c.loc['b','err+'], float(entry['amp']['dres+'][1].get()), 
                        err_msg = 'T1 err+ not equal', decimal=5)
    assert_almost_equal(out_c.loc['b','err-'], float(entry['amp']['dres-'][1].get()), 
                        err_msg = 'amp err- not equal', decimal=5)
    
@with_bfit
def test_1f_1run(b=None, fittab=None, fetchtab=None):
    """
        Test constraining two variables in a single 1f run to be equal
    """
    
    # get 1f data
    fetchtab.remove_all()
    fetchtab.run.set('40037')
    fetchtab.get_data()
    fittab.n_component.set(2)
    fittab.populate()

    # get constrained functions
    constr = popup_fit_constraints(b)
    constr.entry.insert('1.0', 'fwhm_0 = a\nfwhm_1 = a')
    constr.get_input()
    constr.do_fit()
    



# test the global fitter 
# Derek Fujimoto
# Feb 2021

from bfit.fitting.global_fitter import global_fitter
from numpy.testing import *
import numpy as np

# make data sets to fit
fn = [lambda x, a, b: a*x + b, lambda x, a, b: a*x + b]
x = [np.arange(10), np.arange(10)]
y = [x[0]*5+1, x[1]*5+8]
dy = [np.random.rand(10), np.random.rand(10)]
shared = [True, False]

# test function broadcasting
def test_constructor_function_broad():
    gf = global_fitter(fn[0], x, y, dy, shared=shared)
    assert_equal(len(gf.fn), 2, "global fitter function broadcasting length")
    assert_equal(gf.fn[0], gf.fn[1], "global fitter function broadcasting id")

    # test n 
    assert_equal(gf.npar, 2, "global fitter number parameter detection")
    assert_equal(gf.nsets, 2, "global fitter number data sets detection")

# test fixed broadcasting
def test_constructor_fixed_broad():
    gf = global_fitter(fn, x, y, dy, shared=shared, fixed=[False, True])
    assert_equal(gf.fixed[0,1], True, "global fitter fixed broadcasting 1")
    assert_equal(gf.fixed[1,1], True, "global fitter fixed broadcasting 2")
    
    gf = global_fitter(fn, x, y, dy, shared=shared, fixed=[[False, True],[False, False]])
    assert_equal(gf.fixed[0,1], True, "global fitter fixed assignment 1")
    assert_equal(gf.fixed[1,1], False, "global fitter fixed assignment 2")
    
def test_fitting_trf():
    
    gf = global_fitter(fn, x, y, dy, shared=shared)
    gf.fit(minimiser='trf')
    par, std_l, std_h, cov = gf.get_par()

    assert_almost_equal(par[0, 0], par[1, 0], err_msg = "global fitter curve_fit shared parameter equal")
    assert_almost_equal(par[0, 0], 5, err_msg = "global fitter curve_fit parameter 0 result")
    assert_almost_equal(par[0, 1], 1, err_msg = "global fitter curve_fit parameter 1 result")
    assert_almost_equal(par[1, 1], 8, err_msg = "global fitter curve_fit parameter 2 result")

def test_fitting_migrad():
    
    gf = global_fitter(fn, x, y, dy, shared=shared)
    gf.fit(minimiser='migrad')
    par, std_l, std_h, cov = gf.get_par()
    
    assert_almost_equal(abs(par[0, 0] - par[1, 0]), 0, err_msg = "global fitter migrad shared parameter equal")
    assert_almost_equal(abs(par[0, 0] - 5), 0, err_msg = "global fitter migrad parameter 0 result")
    assert_almost_equal(abs(par[0, 1] - 1), 0, err_msg = "global fitter migrad parameter 1 result")
    assert_almost_equal(abs(par[1, 1] - 8), 0, err_msg = "global fitter migrad parameter 2 result")
    
def test_fitting_minos():
    
    gf = global_fitter(fn, x, y, dy, shared=shared)
    gf.fit(minimiser='minos')
    par, std_l, std_h, cov = gf.get_par()
    
    assert_almost_equal(abs(par[0, 0] - par[1, 0]), 0, err_msg = "global fitter minos shared parameter equal")
    assert_almost_equal(abs(par[0, 0] - 5), 0, err_msg = "global fitter minos parameter 0 result")
    assert_almost_equal(abs(par[0, 1] - 1), 0, err_msg = "global fitter minos parameter 1 result")
    assert_almost_equal(abs(par[1, 1] - 8), 0, err_msg = "global fitter minos parameter 2 result")
    
    assert_equal(len(gf.par), 3, "global fitter internal flattened parameter array length")
    
def test_fitting_fixed_trf():
    
    gf = global_fitter(fn, x, y, dy, shared=shared, fixed=[False, True])
    gf.fit(minimiser='trf')
    par, std_l, std_h, cov = gf.get_par()
    
    assert_almost_equal(abs(par[0, 0] - par[1, 0]), 0, err_msg = "global fitter curve_fit shared parameter equal")
    assert_almost_equal(abs(par[1, 1] - 1), 0, err_msg = "global fitter curve_fit parameter fixed result")

def test_fitting_fixed_migrad():
    
    gf = global_fitter(fn, x, y, dy, shared=shared, fixed=[False, True])    
    gf.fit(minimiser='migrad')
    par, std_l, std_h, cov = gf.get_par()
    
    assert_almost_equal(abs(par[0, 0] - par[1, 0]), 0, err_msg = "global fitter migrad shared parameter equal")
    assert_almost_equal(abs(par[1, 1] - 1), 0, err_msg = "global fitter migrad parameter fixed result")
    
def test_fitting_fixed_minos():
    
    gf = global_fitter(fn, x, y, dy, shared=shared, fixed=[False, True])
    gf.fit(minimiser='minos')
    par, std_l, std_h, cov = gf.get_par()
    
    assert_almost_equal(abs(par[0, 0] - par[1, 0]), 0, err_msg = "global fitter minos shared parameter equal")
    assert_almost_equal(abs(par[1, 1] - 1), 0, err_msg = "global fitter minos parameter fixed result")
    
    assert_equal(len(gf.par), 1, "global fitter internal flattened parameter array length with fixed values")
    
def test_fitting_bounds():
    
    gf = global_fitter(fn, x, y, dy, shared=shared)
    gf.fit(minimiser='trf', bounds = [10,11])
    par, std_l, std_h, cov = gf.get_par()
    
    par = np.concatenate(par)
    
    if any(10 > par) and any(par > 11):
        raise AssertionError('Failed: global fitter bounds assignment for format [int int]')
    
    gf.fit(minimiser='trf', bounds = [[10,10.5],11])
    par, std_l, std_h, cov = gf.get_par()
    
    if any(10 > par[:,0]) and any(par[:,0] > 11) and any(10.5 > par[:,1]) and any(par[:,1] > 11):
        print(par)
        raise AssertionError('Failed: global fitter bounds assignment for format [list int]')
    
    gf.fit(minimiser='trf', bounds = [[10,10.5],[10.1,10.6]])
    par, std_l, std_h, cov = gf.get_par()
    
    if any(10 > par[:,0]) and any(par[:,0] > 10.1) and any(10.5 > par[:,1]) and any(par[:,1] > 10.6):
        raise AssertionError('Failed: global fitter bounds assignment for format [list list]')
    
    gf.fit(minimiser='trf', bounds = [[[10,10.5],[10.1,10.6]],[[12,12.5],[12.1,12.6]]])
    par, std_l, std_h, cov = gf.get_par()

    if 10 > par[0,0] > 10.1 and 10.5 > par[0,1] > 10.6 and \
       10 > par[1,0] > 10.1 and 12.5 > par[1,1] > 12.6:
        raise AssertionError('Failed: global fitter bounds assignment for format [[list list]]')
    
def test_chi2():
    
    gf = global_fitter(fn, x, y, dy, shared=shared)
    gf.fit(minimiser='trf')
    
    # test chi2 calc
    gchi, chi = gf.get_chi()
    
    assert(gchi > 0), 'Failed: global fitter chisquared calculation error'
    assert(all(chi > 0)), 'Failed: global fitter chisquared calculation error'
    
